from sentinelsat.sentinel import SentinelAPI # pip install sentinelsat (https://pypi.org/project/sentinelsat/)
from sentinelsat.sentinel import read_geojson
from sentinelsat.sentinel import geojson_to_wkt
from sentinelsat.sentinel import geojson
import requests
from Cryptodome import Random # pip install pycryptodomex (https://pypi.org/project/pycryptodomex/)
from Cryptodome.Cipher import AES
import base64
import hashlib
from collections import OrderedDict
import glob
import os
from process_s2 import extractBands
from osgeo import gdal # Windows: install gdal from WHL file: https://www.lfd.uci.edu/~gohlke/pythonlibs/ ; Linux: pip install gdal
from gdal import gdalconst
import numpy as np
from configuration import Configuration # all user-adjustable parameters can be changed in configuration.py

bs = AES.block_size
key = 'xdfgfces,kghceilgrhseklh'
key = hashlib.sha256(key.encode()).digest()

def pad(s):
    return s + (bs - len(s) % bs) * chr(bs - len(s) % bs)

def unpad(s):
    return s[:-ord(s[len(s)-1:])]

def encrypt(raw):
    raw = pad(raw)
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return base64.b64encode(iv + cipher.encrypt(raw.encode())).decode('utf-8')

def decrypt(enc):
    enc = base64.b64decode(enc)
    iv = enc[:AES.block_size]
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return unpad(cipher.decrypt(enc[AES.block_size:])).decode('utf-8')

def printMeta(product_values, items):
    largest = 0
    for el in items:
        if len(el) > largest:
            largest = len(el);
            
    for el in items:
        print(el.ljust(largest) + ': ' + product_values[el])

def extents(aoi_geojson):
    coords = geojson.utils.coords(aoi_geojson)
    xmin = -1
    xmax = -1
    ymin = -1
    ymax = -1
    for coord in coords:
        if xmin == -1:
            xmin = coord[0]
        else:
            xmin = min(xmin,coord[0])
        if xmax == -1:
            xmax = coord[0]
        else:
            xmax = max(xmax,coord[0])                
        if ymin == -1:
            ymin = coord[1]
        else:
            ymin = min(ymin,coord[1])
        if ymax == -1:
            ymax = coord[1]
        else:
            ymax = max(ymax,coord[1])
    return (xmin,ymin,xmax,ymax)

def ext(nrBands):
    if nrBands > 1:
        return '.mpl'
    else:
        return '.mpr'

def patchCsy(filename):
    # change the mis-detected Mercator projection to Pseudo Mercator
    file=open(filename)
    lines = file.readlines()
    file.close()
    lines = [line.replace('Projection=Mercator','Projection=Pseudo Mercator') for line in lines]
    file=open(filename,'w') # overwrite!!
    file.writelines(lines)
    file.flush()
    file.close()

def patchMpr(filename):    
    # change the single-band raster to the ilwis-type "RGB Color Composite 24" that is not supported by GDAL
    file=open(filename)
    lines = file.readlines()
    file.close()
    lines = [line.replace('Domain=value.dom','Domain=Color.dom') for line in lines]
    lines = [line.replace('Range=0.000:255.000:1.000000:offset=0','DomainInfo=Color.dom;Long;color;0;;') for line in lines]
    lines = [line.replace('Type=Byte','Type=Long') for line in lines]
    # lines.append('Format=2') # not needed?
    file=open(filename,'w') # overwrite!!
    file.writelines(lines)
    file.flush()
    file.close()

def writeMpH(filename, dataset, mask):
    band1 = dataset.GetRasterBand(1).ReadAsArray()
    band2 = dataset.GetRasterBand(2).ReadAsArray()
    band3 = dataset.GetRasterBand(3).ReadAsArray()
    #band4 = dataset.GetRasterBand(4).ReadAsArray()
    band4 = np.full(band1.shape, 255, dtype = np.uint8)
    if not mask is None:
        band1[~mask] = 0
        band2[~mask] = 0
        band3[~mask] = 1
        band4[~mask] = 128 # hex 80
        # pattern is 01 00 00 80 little-endian == ILWIS.iUNDEF
    band1 = band1.flatten()
    band2 = band2.flatten()
    band3 = band3.flatten()
    band4 = band4.flatten()
        
    all_bands = np.empty((band1.size + band2.size + band3.size + band4.size), dtype = band1.dtype)
    all_bands[0::4] = band3
    all_bands[1::4] = band2
    all_bands[2::4] = band1
    all_bands[3::4] = band4
    #all_bands = struct.pack('%dB' % len(all_bands), *all_bands)
    all_bands = bytearray(all_bands)
    file = open(filename,'wb') # this will overwrite the existing mp# !!
    file.write(all_bands)
    file.flush()
    file.close()

def writeIlwisCC(filename, dataset, mask):
    # create a dummy band, to write the mpr/mp#/grf/csy with gdal
    driver = gdal.GetDriverByName('MEM')
    dataset_band1 = driver.Create('', dataset.RasterXSize, dataset.RasterYSize, 1, gdal.GDT_Byte)
    dataset_band1.SetGeoTransform(dataset.GetGeoTransform())
    dataset_band1.SetProjection(dataset.GetProjection())
    rasterband = dataset.GetRasterBand(1)
    data = rasterband.ReadAsArray()
    dataset_band1.GetRasterBand(1).WriteArray(data)
    dataset_ilwis = gdal.Translate(filename + '_rgb.mpr', dataset_band1, format='ILWIS')
    dataset_ilwis.FlushCache()
    dataset_ilwis = None
    # Alter the gdal files
    patchCsy(filename + '_rgb.csy')
    patchMpr(filename + '_rgb.mpr')
    writeMpH(filename + '_rgb.mp#', dataset, mask)

def writeIlwis(filename, dataset, mask):
    if not mask is None:
        driver = gdal.GetDriverByName('MEM')
        dataset_masked = driver.Create('', dataset.RasterXSize, dataset.RasterYSize, dataset.RasterCount, gdal.GDT_Byte)
        dataset_masked.SetGeoTransform(dataset.GetGeoTransform())
        dataset_masked.SetProjection(dataset.GetProjection())
        for i in range(0, dataset.RasterCount):
            rasterband = dataset.GetRasterBand(i + 1)
            data = rasterband.ReadAsArray()
            data[~mask] = 255
            dataset_masked.GetRasterBand(i + 1).WriteArray(data)
    else:
        dataset_masked = dataset
    dataset_ilwis = gdal.Translate(filename + ext(dataset_masked.RasterCount), dataset_masked, format='ILWIS')
    dataset_ilwis.FlushCache()

def minMaxPerc(gdalrasterband, perc):
    stats = gdalrasterband.GetStatistics(False, True) # GetStatistics(bApproxOK, bForce)
    vmin, vmax, vmean, vstd = stats
    rasterarray = gdalrasterband.ReadAsArray()
    pixelcounts, bucketvalues = np.histogram(rasterarray, 256, (1, vmax)) # start at 1, to ignore nodata (which is 0)
    totPix = rasterarray.size
    iPerc = perc * totPix / 100
    ix = 0
    pixCum = pixelcounts[ix]    
    while ix < len(pixelcounts) - 1 and pixCum < iPerc:
        ix = ix + 1
        pixCum = pixCum + pixelcounts[ix]
    minVal =bucketvalues[ix]
    ix = len(pixelcounts) - 1
    pixCum = pixelcounts[ix]
    while ix > 0 and pixCum < iPerc:
        ix = ix - 1
        pixCum = pixCum + pixelcounts[ix]
    maxVal = bucketvalues[ix]
    return minVal, maxVal

def stretch(dataset):
    perc = 1 # 1% interval
    minMaxVals = [minMaxPerc(dataset.GetRasterBand(i+1), perc) for i in range(dataset.RasterCount)]
    driver = gdal.GetDriverByName('MEM')
    dataset_stretched = driver.Create('', dataset.RasterXSize, dataset.RasterYSize, dataset.RasterCount, gdal.GDT_Byte)
    dataset_stretched.SetGeoTransform(dataset.GetGeoTransform())
    dataset_stretched.SetProjection(dataset.GetProjection())
    for i in range(dataset.RasterCount):
        minVal = minMaxVals[i][0] # per-band min/max
        maxVal = minMaxVals[i][1]
        rasterband = dataset.GetRasterBand(i+1)
        data = rasterband.ReadAsArray()
        data = 255*(data-minVal)/(maxVal-minVal)
        dataset_stretched.GetRasterBand(i+1).WriteArray(data)
    return dataset_stretched


def getdatestamps(sentinelfiles):
    datestamps = []
    for sentinelfile in sentinelfiles:
        datestamp = sentinelfile.split('_')[2].split('T')[0]
        if not datestamp in datestamps:
            datestamps.append(datestamp)
    return datestamps

def main():
    path = os.path.dirname(__file__)
    if (len(path) > 0):
        path += os.sep

    if len(Configuration.workdir) > 0:
        print('Reading GeoJSON (*.json) files in', Configuration.workdir)
    else:
        print('Reading GeoJSON (*.json) files in current directory.')
    aois = glob.glob(Configuration.workdir + '*.json')
    if len(aois) == 0:
        print('No file found; please provide at least one GeoJSON file with the area-of-interest.')

    # Download section
    if Configuration.download:
        # The credentials encryption is symmetric, thus given this script and the encrypted credentials file, someone can compute the plain user/passwords
        # However substituting the user/pass with their encrypted versions is recommended as it prevents easy access by curious eyes

        credentialsfile = path + 'credentials_encrypted.txt'
        if os.path.exists(credentialsfile):
            credentials = open(credentialsfile, 'r')
            usern_scihub = decrypt(credentials.readline().strip())
            passw_scihub = decrypt(credentials.readline().strip())
            credentials.close()
        else:
            credentialsfile = path + 'credentials.txt'
            if os.path.exists(credentialsfile):
                credentials = open(credentialsfile, 'r')
                usern_scihub = credentials.readline().strip()
                passw_scihub = credentials.readline().strip()
                credentials.close()
                if len(usern_scihub)>0 and len(passw_scihub)>0:
                    print('Edit the file credentials.txt')
                    print('Replace the content with its encrypted version below')
                    print('Then rename the file to credentials_encrypted.txt')
                    print('After this, run this program again for normal download and processing.')
                    print()
                    print(encrypt(usern_scihub))
                    print(encrypt(passw_scihub))
                    print()
                else:
                    print('Empty or incomplete credentials.txt found.')
                    print('Please put your ESA Scihub credentials in this file, in two separate lines, in this order:')
                    print('ESA Scihub user')
                    print('ESA Scihub pass')
                    print('Nothing else should be in the file; also no spaces and no empty lines.')
                    print('When running this program again, you will be asked to encrypt the credentials.')
                return
            else:
                print('Create a file named credentials.txt in the same folder as Main.py and put your ESA Scihub credentials in two separate lines, in this order:')
                print('ESA Scihub user')
                print('ESA Scihub pass')
                print('Nothing else should be in the file; also no spaces and no empty lines.')
                print('When running this program again, you will be asked to encrypt the credentials.')
                return

        api_scihub = SentinelAPI(usern_scihub, passw_scihub, 'https://scihub.copernicus.eu/dhus')

        sentinel2_producttypes = ['S2MSI2A']

        for aoi in aois:
            aoi=aoi.replace('\\','/')
            print('Downloading Sentinel-2 files for', aoi, 'from', Configuration.starttime, 'to', Configuration.endtime)
            aoi_geojson = read_geojson(aoi)
            footprint = geojson_to_wkt(aoi_geojson)
            footprintname = aoi[0:-5] # remove .json

            allProducts = OrderedDict()
            # Sentinel-2
            for producttype in sentinel2_producttypes:
                if Configuration.cloudcovermax < 100:
                    key_values = {
                            'platformname': 'Sentinel-2',
                            'producttype': producttype,
                            'cloudcoverpercentage': (0, Configuration.cloudcovermax), # from 0 til cloudcovermax; stricter --> less images
                            'date': (Configuration.starttime, Configuration.endtime)}
                else: # no cloudcoverpercentage
                    key_values = {
                            'platformname': 'Sentinel-2',
                            'producttype': producttype,
                            'date': (Configuration.starttime, Configuration.endtime)}

                products = api_scihub.query(footprint, **key_values)
                allProducts.update(products)

            print('Scihub archive query for \'' + footprintname + '\' returned ' + str(len(allProducts)) + ' item(s)')

            if not os.path.exists(footprintname):
                os.makedirs(footprintname)
      
            seq = 1
            for product_key, product_values in allProducts.items():
                #printMeta(product_values, ['uuid','title','platformname','instrumentshortname','producttype','footprint'])
                sentinelfile = product_values['filename'][0:-5]
                zipfile = sentinelfile + '.zip'
                print('[' + str(seq) + '/' + str(len(allProducts)) + '] ' + zipfile)
                if not os.path.exists(footprintname + '/' + zipfile) and not os.path.exists(footprintname + '/' + product_values['filename']):
                    if hasattr(api_scihub, 'is_online'):
                        is_online = api_scihub.is_online(product_key) # is_online is only available after certain api version
                        print('online: ', is_online)
                    else:
                        print('SentinelAPI version too old; can\'t test online presence; assumed yes')
                        is_online = True
                    if is_online:
                        print('Downloading...', end='', flush=True)
                        link = product_values['link']
                        try:
                            with requests.get(link, stream=True, auth=(usern_scihub,passw_scihub)) as response:
                                response.raise_for_status()
                                total_length = response.headers.get('content-length')
                                if total_length is not None:
                                    total_length = int(total_length)
                                downloaded_length = 0
                                with open(footprintname + '/' + zipfile, 'wb') as f:
                                    for chunk in response.iter_content(chunk_size=8192):
                                        downloaded_length += len(chunk)
                                        f.write(chunk)
                            print('done.')
                            seq = seq + 1
                        except:
                            print('failed; product not available; skipping.')
                            seq = seq + 1 # the "continue" below will skip the counter increment
                            continue
                    else:
                        if Configuration.offlineTrigger:
                            print('Product is offline; triggering retrieval from the Long Term Archive; please revisit a bit later; this can take from 1 to 24 hrs.')
                            try:
                                api_scihub.download(product_key, footprintname)
                            except:
                                pass
                        else:
                            print('Product is offline')
                        seq = seq + 1 # the "continue" below will skip the counter increment
                        continue
                else:
                    print('Sentinel-2 zipfile already downloaded.')
                    seq = seq + 1

    # Processing section
    if Configuration.process:
        for aoi in aois:
            aoi=aoi.replace('\\','/')
            aoi_geojson = read_geojson(aoi)
            footprint = geojson_to_wkt(aoi_geojson)
            footprintname = aoi[0:-5] # remove .json
            print('Reading Sentinel-2 zipfiles (*.zip) in folder', footprintname)
            zipfiles = glob.glob(footprintname + '/*.zip')
            if len(zipfiles) == 0:
                print('No file found; download some Sentinel-2 zipfiles first.')
            zipfiles = [zipfile.replace('\\','/') for zipfile in zipfiles]
            sentinelfiles = [zipfile[len(footprintname)+1:-4] for zipfile in zipfiles] # no path, no extension
            datestamps = getdatestamps(sentinelfiles)
            for datestamp in datestamps:
                print('Date:', datestamp)
                for band in Configuration.bands:
                    count = 0
                    for sentinelfile in sentinelfiles:
                        datestamp_sentinel = sentinelfile.split('_')[2].split('T')[0]
                        if datestamp == datestamp_sentinel:
                            count = count + 1
                    print('Band: ', band, '; ', count, ' tiles', sep='', flush=True)
                    if Configuration.autoMosaic and os.path.exists(footprintname + '/' + footprintname.split('/')[-1] + '_' + datestamp + '_' + band + '.mpr'):
                        print('Mosaic file', footprintname.split('/')[-1] + '_' + datestamp + '_' + band + '.mpr', 'already exists.')
                        continue
                    elif Configuration.autoMosaic and os.path.exists(footprintname + '/' + footprintname.split('/')[-1] + '_' + datestamp + '_' + band + '.mpl'):
                        print('Mosaic file', footprintname.split('/')[-1] + '_' + datestamp + '_' + band + '.mpl', 'already exists.')
                        continue
                    seq = 1
                    singleOutputBands = {} # only to be used when Configuration.autoMosaic == True
                    for sentinelfile in sentinelfiles:
                        datestamp_sentinel = sentinelfile.split('_')[2].split('T')[0]
                        if datestamp != datestamp_sentinel:
                            continue
                        tile_id = sentinelfile.split('_')[5]
                        if Configuration.autoMosaic:
                            if (seq > 1):
                                print(', ', sep='', end='', flush=True)
                                print(seq, ': ', tile_id, sep='', end='', flush=True)
                            else:
                                print('Tiles: ', seq, ': ', tile_id, sep='', end='', flush=True)
                        else:
                            print('[',seq,'/',count,'] ',tile_id,' --> ', sep='', end='', flush=True)
                        datasets = extractBands(footprintname, sentinelfile, [band]) #, extents(aoi_geojson))
                        for name, dataset in datasets.items():
                            if Configuration.autoMosaic:
                                try:
                                    datestamp = name.split('_')[1].split('T')[0]
                                    product = name.split('_')[2] + '_' + name.split('_')[3]
                                    name = datestamp + '_' + product
                                    if name in singleOutputBands:
                                        singleOutputBands[name] = gdal.Warp('',[singleOutputBands[name],dataset],format='MEM',resampleAlg=gdalconst.GRA_NearestNeighbour)
                                    else:
                                        singleOutputBands[name] = dataset
                                except:
                                    print('Out of memory; reduce area-of-interest, or specify Configuration.autoMosaic=False')
                                    break
                            else:
                                if Configuration.stretched and dataset.RasterCount > 1:
                                    dataset = stretch(dataset)
                                if Configuration.ilwisCC and dataset.RasterCount > 1:
                                    print(name + '_rgb.mpr')
                                    writeIlwisCC(footprintname + '/' + name, dataset, None)
                                else:
                                    print(name + ext(dataset.RasterCount))
                                    writeIlwis(footprintname + '/' + name, dataset, None)
                        seq = seq + 1
                    if Configuration.autoMosaic:
                        print()
                        print('Writing mosaic: ', sep='', end='', flush=True)
                        for name, singleOutputBand in singleOutputBands.items():
                            if Configuration.stretched and singleOutputBand.RasterCount > 1:
                                singleOutputBand = stretch(singleOutputBand)
                            if Configuration.ilwisCC and singleOutputBand.RasterCount > 1:
                                print(footprintname.split('/')[-1] + '_' + name + '_rgb.mpr')
                                name = footprintname + '/' + footprintname.split('/')[-1] + '_' + name
                                writeIlwisCC(name, singleOutputBand, None)
                            else:
                                print(footprintname.split('/')[-1] + '_' + name + ext(singleOutputBand.RasterCount))
                                name = footprintname + '/' + footprintname.split('/')[-1] + '_' + name
                                writeIlwis(name, singleOutputBand, None)
        print('Done!')

# main()
main()
