from zipfile import ZipFile
import xml.etree.ElementTree as xml
from osgeo import gdal
from uuid import uuid4

# sentinelfile = 'S2B_MSIL2A_20200527T113319_N0214_R080_T28PBA_20200527T140657'

def extractBands(path, sentinelfile, bands):
    filename = sentinelfile
    if (filename.startswith('S2')):
        filename = filename.split('_')
        filename = filename[5] + '_' + filename[2]
        bands = [filename + '_' + band for band in bands]

    filename = (path + '/' + sentinelfile + '.zip')
    z = ZipFile(filename)

    msil = z.open(sentinelfile + '.SAFE/MTD_MSIL2A.xml')
    msil = xml.parse(msil).getroot()
    namespace_msil = msil.tag.split('}')[0] + '}'
    filenames = msil.findall(namespace_msil + 'General_Info/Product_Info/Product_Organisation/Granule_List/Granule/IMAGE_FILE')
    datasets = {}
    for name in filenames:
        name = name.text + '.jp2'
        for band in bands:
            if band in name:
                f = z.open(sentinelfile + '.SAFE/' + name)
                mmap_name = "/vsimem/"+str(uuid4())
                gdal.FileFromMemBuffer(mmap_name, f.read())
                dataset=gdal.Open(mmap_name)
                rasterband = dataset.GetRasterBand(1)
                driver = gdal.GetDriverByName('MEM')
                #dataset_out = driver.CreateCopy(name.split('/')[-1][:-4] + '.mpl', dataset)
                dataset_out = driver.Create('', dataset.RasterXSize, dataset.RasterYSize, dataset.RasterCount, rasterband.DataType)
                dataset_out.SetGeoTransform(dataset.GetGeoTransform())
                dataset_out.SetProjection(dataset.GetProjection())
                if dataset.RasterCount > 1:
                    data = rasterband.ReadAsArray()
                    dataset_out.GetRasterBand(3).WriteArray(data)
                    rasterband = dataset.GetRasterBand(2)
                    data = rasterband.ReadAsArray()
                    dataset_out.GetRasterBand(2).WriteArray(data)
                    rasterband = dataset.GetRasterBand(3)
                    data = rasterband.ReadAsArray()
                    dataset_out.GetRasterBand(1).WriteArray(data)
                else:
                    data = rasterband.ReadAsArray()
                    dataset_out.GetRasterBand(1).WriteArray(data)
                dataset_out.FlushCache() # Commit
                dataset = None
                gdal.Unlink(mmap_name)
                name = name.split('/')[-1][:-4]
                datasets[name] = dataset_out
    return datasets
